#!/usr/bin/env python3
#
# Copyright (c) 2022 Jan Behrens
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import random
import math

# helper function, which re-distributes counts such that they do not exceed
# nearest integers of the actual target values
def catpick_distribute(targets, lows, highs, high_prob):
    # calculate boundaries based on targets
    low_bounds = {}
    high_bounds = {}
    for cat, target in targets.items():
        low_bounds[cat] = math.floor(target)
        high_bounds[cat] = math.ceil(target)
    # categories which can still be adjusted
    open_cats = set(targets)
    while True:
        # errors (positive numbers) introduced in lows or highs
        low_error = 0
        high_error = 0
        # categories to be removed from open_cats after iteration
        cats_to_close = []
        # limit counts in categories which may still be adjusted
        # and remember errors caused
        for cat in open_cats:
            error = low_bounds[cat] - lows[cat]
            if error > 0:
                lows[cat] = low_bounds[cat]
                low_error += error
                other_error = error * ((1-high_prob) / high_prob)
                highs[cat] -= other_error
                high_error += other_error
                cats_to_close.append(cat)
                continue
            error = highs[cat] - high_bounds[cat]
            if error > 0:
                highs[cat] = high_bounds[cat]
                high_error += error
                other_error = error * (high_prob / (1-high_prob))
                lows[cat] += other_error
                low_error += other_error
                cats_to_close.append(cat)
        # mark categories which were force-limited as done
        for cat in cats_to_close:
            open_cats.remove(cat)
        # stop when all categories are done
        if not open_cats: break
        # also stop when there is no more error to be distributed
        if low_error == 0 and high_error == 0: break
        # distribute error
        low_total = 0
        high_total = 0
        for cat in open_cats:
            low_total += lows[cat]
            high_total += highs[cat]
        for cat in open_cats:
            if low_total: lows[cat] -= low_error * lows[cat] / low_total
            if high_total: highs[cat] += high_error * highs[cat] / high_total

# recursive helper function
def catpick_recursive(
    memory, trivial_count, parents, category_mapper, pool, count, round_up
):

    # check illegal count
    if not count >= 0: raise ValueError("count not zero or positive")

    # perform rounding and calculate rounding probability
    count_floor = math.floor(count)
    count_ceil = math.ceil(count)
    count_frac = count % 1
    count_rounded = count_ceil if round_up else count_floor

    # increment memorized category counts
    memory[("cat", parents)] = memory.get(("cat", parents), 0) + count_rounded

    # avoid division by zero
    if count_rounded == 0: return []

    # determine hierarchy level
    level = len(parents)

    # category of each item in pool and number of items per category
    item_cats = []
    totals = {}
    for item in pool:
        cat = category_mapper(level, item)
        item_cats.append(cat)
        totals[cat] = totals.get(cat, 0) + 1

    # set of all categories
    cats = set(totals)

    # total count of pool
    total = len(pool)

    # if all categories are None or if parent category is small enough,
    # simply pick a random number of items
    if total <= trivial_count or cats == set((None,)):
        # try to pick those items which were chosen least often previously
        used_counts = []
        for item in pool:
            used_counts.append(memory.get(("item", item), 0))
        minimum_used_count = min(used_counts)
        preferred = []
        non_preferred = []
        for item, used_count in zip(pool, used_counts):
            if used_count == minimum_used_count:
                preferred.append(item)
            else:
                non_preferred.append(item)
        preferred = random.sample(
            preferred,
            min(len(preferred), count_rounded)
        )
        non_preferred = random.sample(
            non_preferred,
            max(0, count_rounded - len(preferred))
        )
        subpool = preferred + non_preferred
        # increment memorized item counts
        for item in subpool:
            memory[("item", item)] = memory.get(("item", item), 0) + 1
        # return picked items
        return subpool

    # counts to pick per category
    # (real target, and counts in case of rounding count down or up)
    targets, lows, highs = {}, {}, {}
    for cat in cats:
        target = count * totals[cat] / total
        targets[cat] = target
        lows[cat] = target * count_floor / count
        highs[cat] = target * count_ceil / count

    # limit counts to nearest integer
    catpick_distribute(targets, lows, highs, count_frac)

    # target counts for categories based on demanded rounding
    if round_up: targets = highs
    else: targets = lows

    # round target counts to integers
    rounded_counts = {}
    done_counter = 0
    remaining_targets = dict(targets)
    while done_counter < count_rounded:
        cat, target = remaining_targets.popitem()
        if not remaining_targets:
            rounded_counts[cat] = count_rounded - done_counter
            break
        round_up_error = 1 - (target % 1)
        round_down_error = target % 1
        remaining_total = 0
        for remaining_cat, remaining_target in remaining_targets.items():
            remaining_total += remaining_target
        lows, highs = {}, {}
        for remaining_cat, remaining_target in remaining_targets.items():
            if remaining_total > 0:
                lows[remaining_cat] = (
                    remaining_target -
                    round_up_error * remaining_target / remaining_total
                )
                highs[remaining_cat] = (
                    remaining_target +
                    round_down_error * remaining_target / remaining_total
                )
            else:
                lows[remaining_cat] = math.floor(remaining_target + 0.5)
                highs[remaining_cat] = math.floor(remaining_target + 0.5)
        catpick_distribute(remaining_targets, lows, highs, round_up_error)
        if round_up_error > random.random():
            remaining_targets = highs
            rounded_count = math.floor(target)
        else:
            remaining_targets = lows
            rounded_count = math.ceil(target)
        rounded_counts[cat] = rounded_count
        done_counter += rounded_count
    for cat in remaining_targets:
        rounded_counts[cat] = 0

    # consider previous runs and balance out
    aims_total = memory[("cat", parents)] # includes count_rounded already
    errors = {}
    for cat in cats:
        errors[cat] = (
            memory.get(("cat", tuple(parents) + (cat,)), 0)
            + rounded_counts[cat]
            - aims_total * totals[cat] / total
        )
    while True:
        low_errors = {}
        high_errors = {}
        for cat in cats:
            if rounded_counts[cat] + 1 <= math.ceil(targets[cat]):
                low_errors[cat] = errors[cat]
            if rounded_counts[cat] - 1 >= math.floor(targets[cat]):
                high_errors[cat] = errors[cat]
        if not low_errors or not high_errors:
            break
        min_error = min(low_errors.values())
        max_error = max(high_errors.values())
        if min_error >= -2.0 or max_error <= 2.0:
            break
        low_cats = []
        high_cats = []
        for cat, error in low_errors.items():
            if error == min_error:
                low_cats.append(cat)
        for cat, error in high_errors.items():
            if error == max_error:
                high_cats.append(cat)
        low_cat = random.choice(low_cats)
        high_cat = random.choice(high_cats)
        rounded_counts[low_cat] += 1
        rounded_counts[high_cat] -= 1
        errors[low_cat] += 1
        errors[high_cat] -= 1

    # all picked items
    result = []

    # for each category
    for cat in cats:

        # items in category
        subpool = []
        for item, item_cat in zip(pool, item_cats):
            if item_cat == cat:
                subpool.append(item)

        # pick items from category
        result += catpick_recursive(
            memory,
            trivial_count,
            tuple(parents) + (cat,),
            category_mapper,
            subpool,
            targets[cat],
            rounded_counts[cat] > targets[cat]
        )

    # return all picked items
    return result

# picks a certain count of items from a pool and ensures random order,
# where category_mapper maps a level and an item to a category
def catpick(memory, category_mapper, pool, count):
    result = catpick_recursive(
        memory,
        math.floor(len(pool) / math.ceil(count)),
        (),
        category_mapper,
        pool,
        count,
        count % 1 > random.random()
    )
    random.shuffle(result)
    return result

# test case to show that program works correctly
if __name__ == "__main__":

    def rounding_variance(target):
        frac = target % 1
        compl = 1 - frac
        return frac*frac*compl + compl*compl*frac

    def assert_eq(left, right):
        if not abs(left - right) < 1e-12:
            raise AssertionError("testing assertion failed")
    def assert_range(value, low, high):
        if not (value >= low and value <= high):
            print(value, low, high)
            raise AssertionError("testing assertion failed")
    def assert_average(rounds, target, result):
        average = rounds * target
        stddev = math.sqrt(rounds * rounding_variance(target))
        allowed = 6 * stddev
        assert_range(result, average - allowed, average + allowed)

    def category_mapper(level, item):
        result = item[level:level+1]
        if result == "" or result == "-": return None
        return result

    for i in range(1, 1+10):
        batch = i % 2 == 0
        if not batch:
            print("Running test #" + str(i) + " of 10 (independent)...")
        else:
            print("Running test #" + str(i) + " of 10 (batch operation)...")
        target_a1 = random.randint(1, 30)
        target_a2 = random.randint(1, 30)
        target_a3 = random.randint(1, 30)
        target_b1 = random.randint(1, 12)
        target_b2 = random.randint(1, 12)
        target_b3 = random.randint(1, 12)
        target_c1 = random.randint(1, 6)
        target_c2 = random.randint(1, 6)
        target_c3 = random.randint(1, 6)
        target_a = target_a1 + target_a2 + target_a3
        target_b = target_b1 + target_b2 + target_b3
        target_c = target_c1 + target_c2 + target_c3
        count = target_a + target_b + target_c
        pool = []
        for q in range(len(pool)+1, len(pool)+1+target_a1):
            pool.append("A1-" + str(q))
        for q in range(len(pool)+1, len(pool)+1+target_a2):
            pool.append("A2-" + str(q))
        for q in range(len(pool)+1, len(pool)+1+target_a3):
            pool.append("A3-" + str(q))
        for q in range(len(pool)+1, len(pool)+1+target_b1):
            pool.append("B1-" + str(q))
        for q in range(len(pool)+1, len(pool)+1+target_b2):
            pool.append("B2-" + str(q))
        for q in range(len(pool)+1, len(pool)+1+target_b3):
            pool.append("B3-" + str(q))
        for q in range(len(pool)+1, len(pool)+1+target_c1):
            pool.append("C1-" + str(q))
        for q in range(len(pool)+1, len(pool)+1+target_c2):
            pool.append("C2-" + str(q))
        for q in range(len(pool)+1, len(pool)+1+target_c3):
            pool.append("C3-" + str(q))
        if i % 2 == 1:
            random.shuffle(pool)
        sample_count = random.randint(1, random.randint(1, count))
        target_a1 *= sample_count / count
        target_a2 *= sample_count / count
        target_a3 *= sample_count / count
        target_b1 *= sample_count / count
        target_b2 *= sample_count / count
        target_b3 *= sample_count / count
        target_c1 *= sample_count / count
        target_c2 *= sample_count / count
        target_c3 *= sample_count / count
        target_a *= sample_count / count
        target_b *= sample_count / count
        target_c *= sample_count / count
        target_a1_floor = math.floor(target_a1)
        target_a1_ceil = math.ceil(target_a1)
        target_a2_floor = math.floor(target_a2)
        target_a2_ceil = math.ceil(target_a2)
        target_a3_floor = math.floor(target_a3)
        target_a3_ceil = math.ceil(target_a3)
        target_b1_floor = math.floor(target_b1)
        target_b1_ceil = math.ceil(target_b1)
        target_b2_floor = math.floor(target_b2)
        target_b2_ceil = math.ceil(target_b2)
        target_b3_floor = math.floor(target_b3)
        target_b3_ceil = math.ceil(target_b3)
        target_c1_floor = math.floor(target_c1)
        target_c1_ceil = math.ceil(target_c1)
        target_c2_floor = math.floor(target_c2)
        target_c2_ceil = math.ceil(target_c2)
        target_c3_floor = math.floor(target_c3)
        target_c3_ceil = math.ceil(target_c3)
        target_a_floor = math.floor(target_a)
        target_a_ceil = math.ceil(target_a)
        target_b_floor = math.floor(target_b)
        target_b_ceil = math.ceil(target_b)
        target_c_floor = math.floor(target_c)
        target_c_ceil = math.ceil(target_c)
        rounds = 30000
        sum_a1, sum_a2, sum_a3 = 0, 0, 0
        sum_b1, sum_b2, sum_b3 = 0, 0, 0
        sum_c1, sum_c2, sum_c3 = 0, 0, 0
        sum_a, sum_b, sum_c = 0, 0, 0
        memory = {}
        for j in range(rounds):
            if j % 5 == 0 or not batch:
                memory = {}
            samples = catpick(memory, category_mapper, pool, sample_count)
            count_a1 = len(list(filter(lambda x: x[0:2] == "A1", samples)))
            count_a2 = len(list(filter(lambda x: x[0:2] == "A2", samples)))
            count_a3 = len(list(filter(lambda x: x[0:2] == "A3", samples)))
            count_b1 = len(list(filter(lambda x: x[0:2] == "B1", samples)))
            count_b2 = len(list(filter(lambda x: x[0:2] == "B2", samples)))
            count_b3 = len(list(filter(lambda x: x[0:2] == "B3", samples)))
            count_c1 = len(list(filter(lambda x: x[0:2] == "C1", samples)))
            count_c2 = len(list(filter(lambda x: x[0:2] == "C2", samples)))
            count_c3 = len(list(filter(lambda x: x[0:2] == "C3", samples)))
            count_a = count_a1 + count_a2 + count_a3
            count_b = count_b1 + count_b2 + count_b3
            count_c = count_c1 + count_c2 + count_c3
            assert_eq(count_a + count_b + count_c, sample_count)
            assert_range(count_a1, target_a1_floor, target_a1_ceil)
            assert_range(count_a2, target_a2_floor, target_a2_ceil)
            assert_range(count_a3, target_a3_floor, target_a3_ceil)
            assert_range(count_b1, target_b1_floor, target_b1_ceil)
            assert_range(count_b2, target_b2_floor, target_b2_ceil)
            assert_range(count_b3, target_b3_floor, target_b3_ceil)
            assert_range(count_c1, target_c1_floor, target_c1_ceil)
            assert_range(count_c2, target_c2_floor, target_c2_ceil)
            assert_range(count_c3, target_c3_floor, target_c3_ceil)
            assert_range(count_a, target_a_floor, target_a_ceil)
            assert_range(count_b, target_b_floor, target_b_ceil)
            assert_range(count_c, target_c_floor, target_c_ceil)
            sum_a1 += count_a1; sum_a2 += count_a2; sum_a3 += count_a3
            sum_b1 += count_b1; sum_b2 += count_b2; sum_b3 += count_b3
            sum_c1 += count_c1; sum_c2 += count_c2; sum_c3 += count_c3
        sum_a = sum_a1 + sum_a2 + sum_a3
        sum_b = sum_b1 + sum_b2 + sum_b3
        sum_c = sum_c1 + sum_c2 + sum_c3
        assert_average(rounds, target_a1, sum_a1)
        assert_average(rounds, target_a2, sum_a2)
        assert_average(rounds, target_a3, sum_a3)
        assert_average(rounds, target_b1, sum_b1)
        assert_average(rounds, target_b2, sum_b2)
        assert_average(rounds, target_b3, sum_b3)
        assert_average(rounds, target_c1, sum_c1)
        assert_average(rounds, target_c2, sum_c2)
        assert_average(rounds, target_c3, sum_c3)
        assert_average(rounds, target_a, sum_a)
        assert_average(rounds, target_b, sum_b)
        assert_average(rounds, target_c, sum_c)
    print("All tests passed.")
